/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * Author: Anne Rogers, 1994
 * This code was derived from the Split-C code for the same algorithm
 * written and copyrighted by Steve Lumetta, University of California at
 * Berkeley.
 */

#include <stdio.h>
#include <cm/cmmd.h>

void print_graph();

typedef struct node {
  double value;
  int to_length;
  int *to_proc;
  int *to_num;
  int from_length;
  int from_next;
  double **from_values_p;
  double *from_coeffs;
} node_t;



int PROCS, MYPROC;

node_t **h_table, **e_table;
double *gv;

int n_nodes = 1000, d_nodes = 10, local_p = 80, dist_span = 1;
int n_h_nodes, n_e_nodes;

int *h_msgs_from, *e_msgs_from, *start_from, total_from;
int *h_msgs_to, *e_msgs_to, *start_to, total_to;

int *h_sc, *h_rc;
int *e_sc, *e_rc;

double *transfer_buffer;
double **e_trans_ptr, **h_trans_ptr;

/* These constants determine the nature of the random graph. */
#define NUM_H_NODES n_nodes /* 100 */
#define H_NODE_DEGREE d_nodes /* 5 */
#define H_PERCENT_LOCAL local_p /* 0 */

#define NUM_E_NODES n_nodes /* 100 */
#define E_NODE_DEGREE d_nodes /* 5 */
#define E_PERCENT_LOCAL local_p /* 100 */

#define DIST_SPAN dist_span /* 6 */


#define H_COUNT 1
#define E_COUNT 2
#define H_TAG 3
#define E_TAG 4
#define H_GV 5
#define E_GV 6
#define H_VAL 7
#define E_VAL 8

int bytes_allocated=0;
/*********** AUXILIARY  FUNTIONS *****************/
void *mymalloc(int size)
{
  if (size == 0) {
    printf("Calling Malloc with 0\n");
    return NULL;
  }
  else {
    bytes_allocated += size;
    return (void *) malloc(size);
  }
}

/* check_percent returns 1 chance percent of the time--otherwise it 
   returns 0. */
int check_percent(int chance)
{
	return (((random()/128)%100) < chance);
}

/* gen_number generates a random number between 0 and limit - 1. */
int gen_number(int limit)
{
	return (random()/128)%limit;
}

/* generates a number between -(limit - 1) and (limit - 1) excluding 0 */
int gen_signed_number(int limit)
{
	int num = gen_number (2*limit);

	return (num < limit ? (num - limit) : (num - limit + 1));
}


/* generate a random double */
double dbl_random()
{
  return (double)random()/1e9;
}


/* #define assert(index,range) \
    if ((index) >= (range)) \
      printf("%d Index out of range (%x, %d)\n", MYPROC, (index), (range));
*/
/*
void assert(int index, int range)
{  if ((index) >= range)
     printf("%d Index out of range\n", MYPROC);
 }
*/   

#define assert(index, range)

#define mem_check(p,n)                 \
  if ((p == NULL)) {                   \
   printf("Out of memory at %d\n", n);  \
   exit(-1);  }

#define channel_check(c)                \
  if ((c) == CMMD_ERRVAL) {             \
   printf("Out of channels \n");        \
   exit(-2); }
			      



/** BG aux **/

int *count_incoming(node_t **table_out, node_t **table_in, int out_nodes, 
		    int degree, int *incoming_count, double **trans_ptr)
{
  node_t *cn;
  int i, j, k;
  int *count;
  int *to_edges_in;
  int *to_edges_out;
  int in_size, out_size;

  count = (int *) mymalloc(PROCS * sizeof(int));
  mem_check(count, 1);
  to_edges_in = (int *) mymalloc(total_from * sizeof(int));
  mem_check(to_edges_in, 2);
  in_size = total_from*sizeof(int);
  to_edges_out = (int *) mymalloc(total_to * sizeof(int));
  mem_check(to_edges_out, 3);
  out_size = total_to*sizeof(int);



  for (i=0; i < PROCS; i++)
    count[i] = 0;

  /* count "local" incoming edges */
  for (i = 0; i < out_nodes;  i++) {
    cn = table_out[i];
    for (j=0; j < degree; j++) {
      int p = cn->to_proc[j];
      if (p == MYPROC) {
	table_in[cn->to_num[j]]->from_length++;
	assert(cn->to_num[j], n_nodes);
      }
      else {
	to_edges_out[start_to[p]+count[p]] = cn->to_num[j];
	assert(start_to[p]+count[p], out_size);
	trans_ptr[start_to[p]+count[p]] = &cn->value;
	count[p]++;
	assert(p, PROCS);
      }
    }
  }

    
  for (i=0; i < PROCS; i++) {
    if ((MYPROC == i) || (count[i] == 0))
      continue;
    CMMD_free_mcb(CMMD_send_async(i, H_TAG, &to_edges_out[start_to[i]], 
				  count[i]*sizeof(int), NULL, 0));
  }

  for (i=0; i < PROCS; i++) {
    if ((MYPROC == i) || (incoming_count[i] == 0))
      continue;
    CMMD_receive_block(i, H_TAG, &to_edges_in[start_from[i]], 
		       incoming_count[i]*sizeof(int));
    assert(start_from[i]+incoming_count[i]*sizeof(int)-sizeof(int),
	   in_size);
    for (j=0; j < incoming_count[i]; j++) {
      table_in[to_edges_in[start_from[i]+j]]->from_length++;
      assert(to_edges_in[start_from[i]+j], n_nodes);
      assert(start_from[i]+j, in_size);
    }
  }


  return to_edges_in;

}




/******************** Build Graph ***********************/

void all_initialize_graph()
{ node_t *cn, *dest_p;
  int i, j, k;
  int *h_edges_to, *e_edges_to;
  int curr_from, curr_to;


  /* Allocate H nodes */
  n_h_nodes = NUM_H_NODES;
  h_table = (node_t **) mymalloc(NUM_H_NODES*sizeof(node_t *));
  mem_check(h_table, 4);  

  for (i=0; i < n_h_nodes; i++) {
    cn = h_table[i] = (node_t *) mymalloc(sizeof(node_t));
    mem_check(cn, 5);
    cn->from_length = 0;
    cn->value = dbl_random();
  }

  /* allocate E nodes */
  n_e_nodes = NUM_E_NODES;
  e_table = (node_t **) mymalloc(NUM_E_NODES*sizeof(node_t *));
  mem_check(e_table, 6);

  for (i=0; i < n_e_nodes; i++) {
    cn = e_table[i] = (node_t *) mymalloc(sizeof(node_t));
    cn->from_length = 0;
    cn->value = dbl_random();
  }


  /* initialize msg count arrays */
  h_msgs_to = (int *) mymalloc(PROCS*sizeof(int));
  h_msgs_from = (int *) mymalloc(PROCS*sizeof(int));
  e_msgs_to = (int *) mymalloc(PROCS*sizeof(int));
  e_msgs_from = (int *) mymalloc(PROCS*sizeof(int));
  mem_check(h_msgs_to, 7);
  mem_check(h_msgs_from, 8);
  mem_check(e_msgs_to, 9);
  mem_check(e_msgs_from, 10);

  for (i=0; i < PROCS; i++) {
    h_msgs_to[i] = 0;
    h_msgs_from[i] = 0;
    e_msgs_to[i] = 0;
    e_msgs_from[i] = 0;
  }


  /* Generate H edges for the graph -- keeping track of outgoing
     message counts */
  for (i = 0; i < n_h_nodes; i++) {
    cn = h_table[i];
    cn->to_length = H_NODE_DEGREE;
    cn->to_proc = (int *) mymalloc(H_NODE_DEGREE*sizeof(int));
    cn->to_num = (int *) mymalloc(H_NODE_DEGREE*sizeof(int));
    mem_check(cn->to_proc, 11);
    mem_check(cn->to_num, 12);

    for (j=0; j < H_NODE_DEGREE; j++) {
      int dest_proc, dest_num;
      int cp;
      do {
	cp = check_percent(H_PERCENT_LOCAL);
	dest_proc = ((cp)
		     ? MYPROC  
		     : (MYPROC + PROCS + gen_signed_number(DIST_SPAN)) % PROCS);
	dest_num = gen_number(n_e_nodes);

	for (k = 0; k < j; k++)
	  if ((dest_proc == cn->to_proc[k]) &&
	      (dest_num ==  cn->to_num[k]))
	    break;
      } while (k < j);

      cn->to_proc[j] = dest_proc;
      cn->to_num[j] = dest_num;
      h_msgs_to[dest_proc]++;
    }
  }

  /* reset this to 0, easier than checking each time we
     increment h_msgs_to */
  h_msgs_to[MYPROC] = 0;


  /* Generate E edges for the graph -- keeping track of outgoing
     message counts */
  for (i=0; i < n_e_nodes; i++) {
    cn = e_table[i];
    cn->to_length = E_NODE_DEGREE;
    cn->to_proc = (int *) mymalloc(E_NODE_DEGREE*sizeof(int));
    cn->to_num = (int *) mymalloc(E_NODE_DEGREE*sizeof(int));
    mem_check(cn->to_proc, 13);
    mem_check(cn->to_num, 14);

    for (j=0; j < E_NODE_DEGREE; j++) {
      int dest_proc, dest_num;
      do {
	dest_proc = ((check_percent(E_PERCENT_LOCAL)) 
		     ? MYPROC  
		     : (MYPROC + PROCS + gen_signed_number(DIST_SPAN)) % PROCS);
	dest_num = gen_number(n_h_nodes);

	for (k = 0; k < j; k++)
	  if ((dest_proc == cn->to_proc[k]) &&
	      (dest_num ==  cn->to_num[k]))
	    break;
      } while (k < j);

      cn->to_proc[j] = dest_proc;
      cn->to_num[j] = dest_num;
      e_msgs_to[dest_proc]++;
    }
  }

  /* reset this to 0, easier than checking each time we
     increment e_msgs_to */
  e_msgs_to[MYPROC] = 0;



  
  /* Exchange H message counts. (P-1)*(P-1) messages */
  for (j=0; j < PROCS; j++) {
    if (j == MYPROC) 
      continue;
    CMMD_free_mcb(CMMD_send_async(j, H_COUNT, &h_msgs_to[j], 4, NULL, 0));
  }

  for (i=0; i < PROCS; i++) {
    if (i == MYPROC) 
      continue;
    CMMD_receive_block(i, H_COUNT, &h_msgs_from[i], sizeof(int));
  }
  

  /* Exchange E message counts. (P-1)*(P-1) messages */
  for (j=0; j < PROCS; j++) {
    if (j == MYPROC) continue;
    CMMD_free_mcb(CMMD_send_async(j, E_COUNT, &e_msgs_to[j], sizeof(int), NULL, 0));
  }

  for (i=0; i < PROCS; i++) {
    if (i == MYPROC)
      continue;
    CMMD_receive_block(i, E_COUNT, &e_msgs_from[i], sizeof(int));
  }


  start_from = (int *) malloc(PROCS * sizeof(int));
  start_to  = (int *) malloc(PROCS * sizeof(int));

  curr_from = 0;
  curr_to = 0;
  for (i=0; i < PROCS; i++) {
    int to = (h_msgs_to[i] > e_msgs_to[i]) ? h_msgs_to[i] : e_msgs_to[i];
    int from = (h_msgs_from[i] > e_msgs_from[i]) ? h_msgs_from[i] : e_msgs_from[i];

    start_to[i] = curr_to;
    curr_to += to;
    total_to += to;
    start_from[i] = curr_from;
    curr_from += from;
    total_from += from;
  }


  h_trans_ptr = (double **) mymalloc(total_to*sizeof(double *));
  e_trans_ptr = (double **) mymalloc(total_to*sizeof(double *));
  mem_check(h_trans_ptr, 25);
  mem_check(e_trans_ptr, 26);

  h_edges_to = count_incoming(h_table, e_table, n_h_nodes, H_NODE_DEGREE, 
			      h_msgs_from, h_trans_ptr);
  e_edges_to = count_incoming(e_table, h_table, n_e_nodes, E_NODE_DEGREE,
			      e_msgs_from, e_trans_ptr);
  

  /* Allocate space for E from nodes */
  for (i=0; i < n_e_nodes; i++) {
    cn = e_table[i];
    cn->from_next=0;
    if (cn->from_length == 0) continue;
    cn->from_values_p = (double **) mymalloc(cn->from_length*sizeof(double *));
    cn->from_coeffs = (double *) mymalloc(cn->from_length*sizeof(double));
    mem_check(cn->from_values_p, 15);
    mem_check(cn->from_coeffs, 16);
  }

  /* allocate space for ghost nodes */
  gv = (double *) mymalloc (total_from  * sizeof(double));
  mem_check(gv, 17);

  /* create local pointers */
  for (i=0; i < n_h_nodes; i++) {
    cn = h_table[i];
    for (j=0; j < H_NODE_DEGREE; j++) {
      if (cn->to_proc[j] == MYPROC) {
	dest_p = e_table[cn->to_num[j]];
	assert(cn->to_num[j], n_e_nodes);
	dest_p->from_values_p[dest_p->from_next] = &cn->value;
	assert(dest_p->from_next, dest_p->from_length);
	dest_p->from_coeffs[dest_p->from_next] = dbl_random();
	dest_p->from_next += 1;
      }
    }
  }

  /* create H->E ghost pointers */
  for (i=0; i < PROCS; i++) {
    int start = start_from[i];
    if ((MYPROC == i) || (h_msgs_from[i] == 0))
      continue;
  
    for (j=0; j < h_msgs_from[i]; j++) {
      dest_p = e_table[h_edges_to[start+j]];
      assert(start+j, total_from);
      assert(h_edges_to[start+j], n_e_nodes);
      dest_p->from_values_p[dest_p->from_next] = &gv[start_from[i]+j];
      assert(dest_p->from_next, dest_p->from_length);
      assert(start_from[i]+j, total_from);
      dest_p->from_coeffs[dest_p->from_next] = dbl_random();
      dest_p->from_next += 1;
    }
  }




  /* Allocate space for E from nodes */
  for (i=0; i < n_h_nodes;  i++) {
    cn = h_table[i];
    cn->from_next = 0;
    if (cn->from_length == 0) continue;
    cn->from_values_p = (double **) mymalloc(cn->from_length*sizeof(double *));
    cn->from_coeffs = (double *) mymalloc(cn->from_length*sizeof(double));
    mem_check(cn->from_values_p, 18);
    mem_check(cn->from_coeffs, 19);
  }



  /* create E->H local pointers */
  for (i=0; i < n_e_nodes; i++)  {
    cn = e_table[i];
    for (j=0; j < E_NODE_DEGREE; j++) {
      if (cn->to_proc[j] == MYPROC) {
	dest_p = h_table[cn->to_num[j]];
	assert(cn->to_num[j], n_h_nodes);
	dest_p->from_values_p[dest_p->from_next] = &cn->value;
	assert(dest_p->from_next, dest_p->from_length);
	dest_p->from_coeffs[dest_p->from_next] = dbl_random();
	dest_p->from_next = dest_p->from_next + 1;
      }
    }
  }



  /* create H->E ghost pointers */
  for (i=0; i < PROCS; i++) {
    int start = start_from[i];
    if ((MYPROC == i) || (e_msgs_from[i] == 0))
      continue;
  
    for (j=0; j < e_msgs_from[i]; j++) {
      dest_p = h_table[e_edges_to[start+j]];
      assert(e_edges_to[start+j], n_h_nodes);
      assert(start+j, total_from);
      dest_p->from_values_p[dest_p->from_next] = &gv[start_from[i]+j];
      assert(dest_p->from_next, dest_p->from_length);
      assert(start_from[i]+j, total_from);
      dest_p->from_coeffs[dest_p->from_next] = dbl_random();
	dest_p->from_next = dest_p->from_next + 1;
    }
  }

  transfer_buffer = (double *) mymalloc(total_to*sizeof(double));
  h_sc = (int *) mymalloc(PROCS * sizeof(int));
  h_rc = (int *) mymalloc(PROCS * sizeof(int));
  e_sc = (int *) mymalloc(PROCS * sizeof(int));
  e_rc = (int *) mymalloc(PROCS * sizeof(int));
  mem_check(transfer_buffer, 19);
  mem_check(h_sc, 20);
  mem_check(h_rc, 21);
  mem_check(e_sc, 22);
  mem_check(e_rc, 23);

  /* set up channels */
  for (i=0; i < PROCS; i++) {
    if (i == MYPROC)
      continue;
    if (h_msgs_to[i] != 0) {
      h_sc[i] = CMMD_open_send_channel(i, H_VAL, &transfer_buffer[start_to[i]],
				       h_msgs_to[i]*sizeof(double));
      if (h_sc[i] == CMMD_ERRVAL)
	printf("Out of channels\n");
    }
    if (h_msgs_from[i] != 0) {
      h_rc[i] = CMMD_open_receive_channel(i, H_VAL, &gv[start_from[i]], 
					  h_msgs_from[i]*sizeof(double));
      if (h_rc[i] == CMMD_ERRVAL)
	printf("Out of channels\n");
    }
    if (e_msgs_to[i] != 0) {
      e_sc[i] = CMMD_open_send_channel(i, E_VAL, &transfer_buffer[start_to[i]],
				       e_msgs_to[i]*sizeof(double));
      if (e_sc[i] == CMMD_ERRVAL)
	printf("out of channels\n");
    }
    if (e_msgs_from[i] != 0) {
      e_rc[i] = CMMD_open_receive_channel(i, E_VAL, &gv[start_from[i]], 
					  e_msgs_from[i]*sizeof(double));
      if (e_rc[i] == CMMD_ERRVAL)
	printf("Out of channels\n");
    }
  }

  

}


/*************** Shuffle Data *********************/

void move_h_data(int first_time)
{ 
  int i, j;
  node_t *cn;

  if (!first_time) {
    for (i=0; i < PROCS; i++) {
      if ((i == MYPROC) || (h_msgs_from[i] == 0))
	continue;
      CMMD_reset_channel(h_rc[i]);
    }
  }


  for (i=0; i < PROCS; i++) {
    int start = start_to[i];
    for (j=0; j < h_msgs_to[i]; j++) 
      transfer_buffer[start+j] = *h_trans_ptr[start+j];
  }

  for (i=0; i < PROCS; i++) {
    if ((i == MYPROC) || (h_msgs_to[i] == 0))
      continue;
    while (!CMMD_is_channel_writable(h_sc[i]))
      continue;
    CMMD_write_channel(h_sc[i]);
  }

  for (i=0; i < PROCS; i++) {
    if ((i == MYPROC) || (h_msgs_from[i] == 0))
      continue;
    while (!CMMD_is_channel_readable(h_rc[i]))
      continue;
  }
  
  
}


void move_e_data(int first_time)
{ 
  int i, j;
  node_t *cn;


  if (!first_time) {
    for (i=0; i < PROCS; i++) {
      if ((i == MYPROC) || (e_msgs_from[i] == 0))
	continue;
      CMMD_reset_channel(e_rc[i]);
    }
  }


  for (i=0; i < PROCS; i++) {
    int start = start_to[i];
    for (j=0; j < e_msgs_to[i]; j++)
      transfer_buffer[start+j] = *e_trans_ptr[start+j];
  }

  for (i=0; i < PROCS; i++) {
    if ((i == MYPROC) || (e_msgs_to[i] == 0))
      continue;
    while (!CMMD_is_channel_writable(e_sc[i]))
      continue;
    CMMD_write_channel(e_sc[i]);
  }

  for (i=0; i < PROCS; i++) {
    if ((i == MYPROC) || (e_msgs_from[i] == 0))
	continue;
    while (!CMMD_is_channel_readable(e_rc[i]))
      continue;
  }
  
  
}


 


/******************* Compute **************************/

void all_compute_e(int first_time)
{ int i, j;
  node_t *cn;

  move_h_data(first_time);
  for (i = 0; i < n_e_nodes; i++) {
    register double cur_value;

    cn = e_table[i];
    cur_value = cn->value;

    for (j=0; j < cn->from_length - 1;)
      {
	register double *p1, *p2;
        register double d1, d2, d3, d4, d5, d6;

        /* do two iterations (unrolled) of the following statement */
        /* cn->value -= (*cn->from_values_p[j])*cn->from_coeffs[j]; */

        p1 = cn->from_values_p[j];
        d1 = cn->from_coeffs[j];
        d2 = *p1;
        p2 = cn->from_values_p[j+1];
        d5 = d1*d2;
        d3 = *p2;
        d4 = cn->from_coeffs[j +1];
        cur_value -= d5;
        d6 = d3*d4;
        j += 2;
        cur_value -= d6;
      }
    if (j == cn->from_length - 1)
      {
        cur_value -= (*cn->from_values_p[j])*cn->from_coeffs[j];
      }
    cn->value = cur_value;
  }
}


void all_compute_h(int first_time)
{ int i, j;
  node_t *cn;

  move_e_data(first_time);
  for (i = 0; i < n_h_nodes; i++) {
    register double cur_value;

    cn = h_table[i];
    cur_value = cn->value;

    for (j=0; j < cn->from_length - 1;)
      {
	register double *p1, *p2;
        register double d1, d2, d3, d4, d5, d6;

        /* do two iterations (unrolled) of the following statement */
        /* cn->value -= (*cn->from_values_p[j])*cn->from_coeffs[j]; */

        p1 = cn->from_values_p[j];
        d1 = cn->from_coeffs[j];
        d2 = *p1;
        p2 = cn->from_values_p[j+1];
        d5 = d1*d2;
        d3 = *p2;
        d4 = cn->from_coeffs[j +1];
        cur_value -= d5;
        d6 = d3*d4;
        j += 2;
        cur_value -= d6;
      }
    if (j == cn->from_length - 1)
      {
        cur_value -= (*cn->from_values_p[j])*cn->from_coeffs[j];
      }
    cn->value = cur_value;
  }
}
    


main(int argc, char *argv[])
{ int t;
  int iters = 50;
  int first = 1;

#ifndef WWT
  CMMD_fset_io_mode(stdout, CMMD_independent);
#endif

  if (argc > 1)  /* Number of nodes */
    sscanf(argv[1], "%d", &n_nodes);
  if (argc > 2)  /* Degree of nodes */
    sscanf(argv[2], "%d", &d_nodes);
  if (argc > 3)  /* Percent of remote links */
    if (sscanf(argv[3], "%d", &local_p))
      local_p = 100 - local_p;
  if (argc > 4)  /* Distribution span */
    sscanf(argv[4], "%d", &dist_span);
  if (argc > 5)  /* Distribution span */
    sscanf(argv[5], "%d", &iters);

  PROCS =  CMMD_partition_size();
  MYPROC = CMMD_self_address();

  if (MYPROC == 0)
    printf("Nodes %d, degree %d, local %d\n", n_nodes, d_nodes, local_p);

  srandom(17*MYPROC);

  if (MYPROC < PROCS) {
    if (MYPROC == 0) {
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);
    }

    all_initialize_graph();

    if (MYPROC == 0)
      printf("bytes allocated = %d\n", bytes_allocated);
#ifndef WWT
/*
    if (MYPROC == 0) {
      CMMD_node_timer_stop(0);
      printf("Init time %f second\n", CMMD_node_timer_elapsed(0));
      CMMD_node_timer_clear(0);
      CMMD_node_timer_start(0);
    }
*/
#endif

    for (t = 0; t < iters; t++) {
      all_compute_e(first);
      all_compute_h(first);
      first = 0;
      if (MYPROC == 0)
	printf("Iteration %d finished\n", t);
    }
    CMMD_sync_with_nodes();
    CMMD_all_msgs_wait();
    if (MYPROC == 0) {
      CMMD_node_timer_stop(0);
      printf("Compute time %f seconds\n",
	     CMMD_node_timer_elapsed(0));
    }
    exit(0);
  }
  else exit(0);

  
}


void print_graph()
{ int i, j;
  node_t *cn;
  int temp = 0;

  if (MYPROC != 0)
    CMMD_receive_block(MYPROC-1, 1000, &temp, 4);

  printf("\n\nMYPROC = %d\n\n", MYPROC);
  /* print H values & edges */
  printf("H values\n");
  for (i=0; i < n_h_nodes; i++) {
    cn = h_table[i];
    printf("%d -- value = %f, from_length = %d, from_next = %d\n", i, 
	   cn->value, cn->from_length, cn->from_next);
    for (j=0; j < H_NODE_DEGREE; j++)
      printf("<%d, %d>   ", cn->to_proc[j], cn->to_num[j]);

    printf("\nCoeffs\n");
    for (j=0; j < cn->from_next; j++)
      printf("%f   ",cn->from_coeffs[j]);
    printf("\nValues \n");
    for (j=0; j < cn->from_next; j++)
      printf("%f   ",*cn->from_values_p[j]);
    printf("\n\n");
    
  }

  /* print E values & edges */
  printf("E information \n");
  for (i=0; i < n_e_nodes; i++) {
    cn = e_table[i];
    printf("%d -- value = %f, from_length = %d, from_next = %d\n", i, cn->value,
	   cn->from_length, cn->from_next);
    for (j=0; j < E_NODE_DEGREE; j++)
      printf("<%d, %d>   ", cn->to_proc[j], cn->to_num[j]);

    printf("\ncoeffs");
    for (j=0; j < cn->from_next; j++)
      printf("%f   ", cn->from_coeffs[j]);
    printf("\nValues\n");
    for (j=0; j < cn->from_next; j++)
      printf("%f   ", *cn->from_values_p[j]);
    printf("\n\n");
    

  }

  if (MYPROC != (PROCS-1))
    CMMD_send_block(MYPROC+1, 1000, &temp, 4);
}


