/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * Author: Satish Chandra, 1994
 */


#include "f2c.h"

/* Fortran Interface function */
void initrandom_(long int *l)
{
  srandom(*l);
}

/*     SRANDOM */
doublereal srandom_(void)
{
    /* System generated locals */
    doublereal ret_val;

    /* Local variables */
    static doublereal r;
    extern integer rand(void);
    integer i;

    i = random();

    r = ((doublereal) i) / (doublereal) 0x7fffffff;
    ret_val = r;
    return ret_val;
} /* srandom_ */

#include <assert.h>
#include <cm/cmmd.h>
#include <math.h>
#include "eventlog.h"

/* Broadcast Functions */

/* Fortran Interface */

int do_broadcast_(int *tag, void *outbuffer, int *nbytes, int *me, int *n)
{
  return DoBroadcast(*tag, outbuffer, *nbytes, *me, *n);
}

int get_broadcast_(int *tag, void *inbuffer, int *nbytes, int *me, int *n)
{
  return GetBroadcast(*tag, inbuffer, *nbytes, *me, *n);
}

int DoBroadcast(int tag, void *outbuffer, int nbytes, int me, int n)
{
  int nlevel, dest, l, l_bit;

  wwt_log_ev(EV_ENTER_BC,0,0,0);

  assert(n >= 2);
#ifndef SSBC
  /* send to leaders of siblings at all levels. Leader at a level
     is the proc whose id has last l bits zero */
  nlevel = (int) log2((double) n);
  for (l = nlevel; l >= 1; l--)
    {
      /* send to flip_l_th_bit and zero last (l - 1) bits */
      l_bit = ((me & (1 << (l - 1))) >> (l-1));
      if (l_bit == 0)
	dest = (me | (1 << (l - 1)));
      else
	dest = (me & ~(1 << (l - 1)));

      dest &= ~((1 << (l-1)) - 1);
      CMMD_free_mcb(CMMD_send_async(dest, tag, outbuffer, nbytes,NULL,0));
    }
  while (!CMMD_all_sends_done());
#else
  for (l = 0;l < n;l++)
    if (l != me)
      CMMD_free_mcb(CMMD_send_async(l,tag,outbuffer,nbytes,NULL,0));
  while (!CMMD_all_sends_done());
#endif
  wwt_log_ev(EV_EXIT_BC,0,0,0);
}
      
int GetBroadcast(int tag, void *inbuffer, int nbytes, int me, int n)
{
  int l, nlevel, k, sender;
  extern void wwt_start_lib_code();
  extern void wwt_end_lib_code();

  wwt_log_ev(EV_ENTER_BC,0,0,0);

  assert(n >= 2);

  nlevel = (int) log2((double) n);
  
  CMMD_receive_block(CMMD_ANY_NODE, tag, inbuffer, nbytes);
#ifndef SSBC
  sender = CMMD_msg_sender();
  
  /* for each level, see if I am the leader of that level, if yes
     send to leaders at all levels less than current */
  for (l = nlevel - 1; l >= 1; l--)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero */
      {
	if ((sender < me) || (sender > (me + (1 << l) - 1)))
	  {
	    for (k = l; k >= 1;k--)
	      {
		CMMD_free_mcb(CMMD_send_async(me + (1 << (k-1)), 
					      tag, inbuffer, nbytes, NULL, 0));
	      }
	    while (!CMMD_all_sends_done());
	    break;
	  }
      }
#endif
  wwt_log_ev(EV_EXIT_BC, 0,0,0);
}

int do_gssum_(int *tag, void *sum, int *nfloats, void *sum2, int *me, int *n)
{
  return DoGssum(*tag, sum, *nfloats, sum2, *me, *n);
}

int DoGssum(int tag, real *sum, int nfloats, real *sum2, int me, int n)
{
  int l, nlevel, k;
  int realsize = 4;
  
  /* Tree algorithm */
  nlevel = (int) log2((double) n);
  
  for (l = 1;l <= nlevel;l++)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero ? */
      {
	CMMD_receive_block(CMMD_ANY_NODE, tag, sum2, realsize * nfloats);
	for (k = 0;k < nfloats;k++)
	  sum[k] += sum2[k];
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	{
	  CMMD_send_block((me & ~(0x1)),tag,sum,realsize*nfloats);
	}
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  {
	    CMMD_send_block((me & ~((1 << l) - 1)),tag, sum, realsize*nfloats);
	  }
  for (l = nlevel;l >= 1;l--)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero */
      {
	CMMD_send_block(me + (1 << (l - 1)),tag,sum, realsize*nfloats);
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	CMMD_receive_block(CMMD_ANY_NODE,tag,sum,realsize*nfloats);
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  CMMD_receive_block(CMMD_ANY_NODE,tag,sum,realsize*nfloats);
}


int do_gshigh_(int *tag, void *sum, int *nfloats, void *sum2, int *me, int *n)
{
  return DoGshigh(*tag, sum, *nfloats, sum2, *me, *n);
}

int DoGshigh(int tag, real *sum, int nfloats, real *sum2, int me, int n)
{
  int l, nlevel, k;
  int realsize = 4;
  
  /* Tree algorithm */
  nlevel = (int) log2((double) n);

  for (l = 1;l <= nlevel;l++)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero ? */
      {
	CMMD_receive_block(CMMD_ANY_NODE, tag, sum2, realsize * nfloats);
	for (k = 0;k < nfloats;k++)
          if (sum[k] < sum2[k])
            sum[k] = sum2[k];
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	{
	  CMMD_send_block((me & ~(0x1)),tag,sum,realsize*nfloats);
	}
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  {
	    CMMD_send_block((me & ~((1 << l) - 1)),tag, sum, realsize*nfloats);
	  }
  for (l = nlevel;l >= 1;l--)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero */
      {
	CMMD_send_block(me + (1 << (l - 1)),tag,sum, realsize*nfloats);
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	{
	  CMMD_receive_block(CMMD_ANY_NODE,tag,sum,realsize*nfloats);
	}
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  {
	    CMMD_receive_block(CMMD_ANY_NODE,tag,sum,realsize*nfloats);
	  }
}

/* Channel Implementation of 1 word broadcast and reduction */

/* data */

/* the up variables are for the "up" phase of the reduction */

#define MAX_SCD 20
#define CHANNEL_1_TAG 9999
#define CHANNEL_2_TAG 8888

#define REAL_ADD     1
#define REAL_MAX     2

int SendBuffer;
int RecvBuffer;

int Up_SendBuffer;
int Up_RecvBuffer[MAX_SCD];

int num_send_scd = 0;
int num_up_recv_scd = 0;

int send_scd[MAX_SCD];
int recv_scd;

int up_send_scd;
int up_recv_scd[MAX_SCD];

/*
int bcast_word;
volatile int got_bcast_word = 0;
volatile int sent_bcast_word = 0;

void ok_got_bcast_word();

void get_bcast_word(int data, int who)
{
  bcast_word = data;
  got_bcast_word = 1;
  CMAML_reply(who, ok_got_bcast_word);
}

void ok_got_bcast_word()
{
  sent_bcast_word = 1;
}
*/
int Init_1_Bcast(int me, int n)
{
  int l, nlevel;

  /* Tree algorithm */
  nlevel = (int) log2((double) n);

  for (l = nlevel;l >= 1;l--)
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero */
      {
	send_scd[num_send_scd] = CMMD_open_send_channel(me+(1<<(l-1)),
							CHANNEL_1_TAG,
							(void *) &SendBuffer,
							4);
	while (CMMD_send_channel_status(send_scd[num_send_scd]) == CMMD_O_send);
	num_send_scd++;
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	{
	  recv_scd = CMMD_open_receive_channel(me-1,
					       CHANNEL_1_TAG,
					       (void *) &RecvBuffer,
					       4);
	  while (CMMD_receive_channel_status(recv_scd) == CMMD_O_receive);
	}
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  {
	    recv_scd = CMMD_open_receive_channel(me-(1<<(l-1)),
						 CHANNEL_1_TAG,
						 (void *) &RecvBuffer,
						 4);
	    while (CMMD_receive_channel_status(recv_scd) == CMMD_O_receive);
	  }
}
#if 0
int Init_1_Upcast(int me, int n)
{
  int l, nlevel;
  
  nlevel = (int) log2((double) n);
  
  for (l = 1;l <= nlevel;l++)
    {
    if ((me & ((1 << l) - 1)) == 0) /* last l bits zero ? */
      {
	up_recv_scd[num_up_recv_scd] = 
	  CMMD_open_receive_channel(me+(1<<(l-1)),
				    CHANNEL_2_TAG,
				    (void *) &(Up_RecvBuffer[num_up_recv_scd]),
				    4);
	while (CMMD_receive_channel_status(up_recv_scd[num_up_recv_scd]) 
	       == CMMD_O_receive);
	num_up_recv_scd++;
      }
    else
      /* last l bits are 1 followed by (l-1) zeroes */
      if (l == 1)
	{
	  up_send_scd = CMMD_open_send_channel(me-1,
					       CHANNEL_2_TAG,
					       (void *) &Up_SendBuffer,
					       4);
	  while (CMMD_send_channel_status(up_send_scd) == CMMD_O_send);
	}
      else
	if (((me & ((1 << (l-1)) - 1)) == 0) && ((me & ((1 << l) - 1)) != 0))
	  {
	    up_send_scd = CMMD_open_send_channel(me-(1<<(l-1)),
						 CHANNEL_2_TAG,
						 (void *) &Up_SendBuffer,
						 4);
	    while (CMMD_send_channel_status(up_send_scd) == CMMD_O_send);
	  }
    CMMD_sync_with_nodes();
  }
}
#endif
int Do_1_Bcast(int *data, int me)
{
  int i;

  wwt_log_ev(EV_ENTER_BC,0,0,0);
  
  if (me != 0)
    {
      CMMD_send_block(0, CHANNEL_1_TAG, data, 4); 
      /* CMAML_request(0,get_bcast_word,*data,me); 
	 while (!sent_bcast_word);
	 sent_bcast_word = 0;
	 */
      
      SendBuffer = *data;
      for (i = 0;i < num_send_scd;i++)
	{
	  while (!CMMD_is_channel_writable(send_scd[i]));
	  CMMD_write_channel(send_scd[i]);
	}
      while (!CMMD_is_channel_readable(recv_scd));
      CMMD_reset_channel(recv_scd);
    }
  else
    {
      SendBuffer = *data;
      
      for (i = 0;i < num_send_scd;i++)
	{
	  while (!CMMD_is_channel_writable(send_scd[i]));
	  CMMD_write_channel(send_scd[i]);
	}
    }

  wwt_log_ev(EV_EXIT_BC,0,0,0);
}

int Get_1_Bcast(int *data, int me)
{
  int i;
  
  wwt_log_ev(EV_ENTER_BC,0,0,0);

  if (me != 0)
    {
      while (!CMMD_is_channel_readable(recv_scd));
      *data = RecvBuffer;
      CMMD_reset_channel(recv_scd);
     
      SendBuffer = *data;
      
      for (i = 0;i < num_send_scd;i++)
	{
	  while (!CMMD_is_channel_writable(send_scd[i]));
	  CMMD_write_channel(send_scd[i]);
	}
    }
  else
    {
      CMMD_receive_block(CMMD_ANY_NODE, CHANNEL_1_TAG, data, 4);
      /*
      while (!got_bcast_word);
      got_bcast_word = 0;
      SendBuffer = bcast_word;
      */

      SendBuffer = *data;
      for (i = 0;i < num_send_scd;i++)
	{
	  while (!CMMD_is_channel_writable(send_scd[i]));
	  CMMD_write_channel(send_scd[i]);
	}
    }
  wwt_log_ev(EV_EXIT_BC,0,0,0);
}

int Do_1_Reduction(void *data, int me, void *data2, int op)
{
  int i;

  wwt_log_ev(EV_ENTER_BC, 0, 0, 0);
#ifndef SBC
  /*
  for (i = 0;i < num_up_recv_scd;i++)
  */
  for (i = num_up_recv_scd - 1; i >= 0;i--)
    {
      while (!CMMD_is_channel_readable(up_recv_scd[i]));
      *((int *) data2) = Up_RecvBuffer[i];
      CMMD_reset_channel(up_recv_scd[i]);

      /*printf("data was %f, data got %f,",*(float *)data,*(float *)data2);*/
      if (op == REAL_ADD)
	*((float *) data) += *((float *) data2);
      else /* op == REAL_MAX */
	if (*((float *) data) < *((float *) data2))
	  *((float *) data) = *((float *) data2);
      /*printf("new data = %f\n",*(float *)data);*/
    }
  if (me != 0)
    {
      Up_SendBuffer = *((int *) data);
      while (!CMMD_is_channel_writable(up_send_scd));
      if (CMMD_write_channel(up_send_scd))
	printf("error here\n");
      /* Get_1_Bcast(data, me); */
      /* GetBroadcast(0, data, 4, me, 32);  */
      CMMD_receive_bc_from_node(data, 4);
    }
  else
    {
      /* Do_1_Bcast(data, me); */
      /* DoBroadcast(0,data,4,me,32); */
      CMMD_bc_to_nodes(data, 4);
    }
  /*printf("after reduction: %f\n",*(float *)data);*/
#endif
  wwt_log_ev(EV_EXIT_BC, 0, 0, 0);
}

/* Fortran Interface */

int Init_1_Upcast(int,int);

int init_1_bcast_(int *me, int *n)
{
  wwt_log_ev(EV_ENTER_BC,0,0,0);
  /* Init_1_Bcast(*me,*n); */
  Init_1_Upcast(*me,*n);
  wwt_log_ev(EV_EXIT_BC,0,0,0);
  return 1;
}

int do_1_bcast_(void *data, int *me)
{
  return Do_1_Bcast(data, *me);
}

int get_1_bcast_(void *data, int *me)
{
  return Get_1_Bcast(data, *me);
}

int do_1_gshigh_(void *data, void *data2, int *me)
{
  return Do_1_Reduction(data,*me,data2,REAL_MAX);
}

/* WARNING: This code assumes 32 procs 
*/

struct tree_data {
  int parent;
  int children[10];
} FibTree[] = { 
  /* 0 */ { -1, {1, 2, 3, 5, 8, 13, 21, -1}},
  /* 1 */ { 0, {4, 6, 9, 14, 22, -1}},
  /* 2 */ { 0, {7, 10, 15, 23, -1}},
  /* 3 */ { 0, {11, 16, 24, -1}},
  /* 4 */ { 1, {12, 17, 25, -1}},
  /* 5 */ { 0, {18, 26, -1}},
  /* 6 */ { 1, {19, 27, -1}},
  /* 7 */ { 2, {20, 28, -1}},
  /* 8 */ { 0, {29, -1}},
  /* 9 */ { 1, {30, -1}},
  /* 10 */ {2, {31, -1}},
  /* 11 */ {3, {-1}},
  /* 12 */ {4, {-1}},
  /* 13 */ {0, {-1}},
  /* 14 */ {1, {-1}},
  /* 15 */ {2, {-1}},
  /* 16 */ {3, {-1}},
  /* 17 */ {4, {-1}},
  /* 18 */ {5, {-1}},
  /* 19 */ {6, {-1}},
  /* 20 */ {7, {-1}},
  /* 21 */ {0, {-1}},
  /* 22 */ {1, {-1}},
  /* 23 */ {2, {-1}},
  /* 24 */ {3, {-1}},
  /* 25 */ {4, {-1}},
  /* 26 */ {5, {-1}},
  /* 27 */ {6, {-1}},
  /* 28 */ {7, {-1}},
  /* 29 */ {8, {-1}},
  /* 30 */ {9, {-1}},
  /* 31 */ {10, {-1}}
	};

int Init_1_Upcast(int me, int n)
{  
  int i, dest, src;

  for (i = 0;;i++)
    {
      dest = FibTree[me].children[i];
      if (dest == -1) break;
      up_recv_scd[num_up_recv_scd] =
	CMMD_open_receive_channel(dest,CHANNEL_2_TAG,
				  (void *) &(Up_RecvBuffer[num_up_recv_scd]),
				  4);
      while (CMMD_receive_channel_status(up_recv_scd[num_up_recv_scd]) 
	     == CMMD_O_receive);
      num_up_recv_scd++;
    }
  src = FibTree[me].parent;
  if (me != 0)
    {
      up_send_scd = CMMD_open_send_channel(src, CHANNEL_2_TAG,
					   (void *) &Up_SendBuffer,
					   4);
      while (CMMD_send_channel_status(up_send_scd) == CMMD_O_send);
    }
  CMMD_sync_with_nodes();
}






