/* Copyright (c) 1994-95 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 */
/*------------------------------------------------------------
   lap.c
  ------------------------------------------------------------
   used modules: 
   geometry.c
  ------------------------------------------------------------
   programmed by Frank Traenkle 
   original Fortran-Version
   by Prof. Sangtae Kim   
  ------------------------------------------------------------
   each processor has subset of rows
   buffered submatrices
   async iteration   
   n particles in 3D-space  
   advanced output
  ------------------------------------------------------------*/

/*
   Deflation Version, solved by Jacobi Iterations
   No Gaussian Quadratures if elements are not near neighbors.
*/

/*----------------------------*/
/* INCLUDEs                   */
/*----------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <cm/cmmd.h>
#include <cm/cmmd/io.h>
#include "my.h"
#include "geometry.h"

/*----------------------------*/
/* DEFINEs                    */
/*----------------------------*/
#define FILENAM_SIZE 40
#define STRING_SIZE 80
#define P_ 32            /* number of processors */
#define M_ 20             /* number of elements per particle */
#define N_ 256           /* number of particles (must be a multiple 
                             of P_) */
#define NBUF_ 240         /* NBUF_ * N_ / P_ = 
                             number of matrices to be buffered on one
                             processor */
#define NM_ (N_ * M_)     /* dimension of solution vector */
#define NMP_ (NM_ / P_)   /* number of rows per processor */
#define NP_ (N_ / P_)     /* number of particles per processor */
/* define DEF_OUTMSG to output temporary message file */
/*
#define DEF_OUTMSG
*/

/* maximum number of send messages */
#define SENDMSG_MAX P_
/* timers */
#define TIMER_TOTAL                  0
#define TIMER_INITIALIZE             1
#define TIMER_STOREELEMENTS          2
#define TIMER_CLOSEGEOMETRY          3
#define TIMER_ADJUSTGEOMETRY         4
#define TIMER_CREATEMATRICES         5
#define TIMER_SOLVE                  6
#define TIMER_CLEANUP                7
#define TIMER_OUTPUT                 8
#define TIMER_COMMUNICATE            9
#define TIMER_RECREATEMATRICES       10
#define TIMER_INITIALIZESOLVE        11
#define TIMER_MSG                    12
#define TIMER_BARRIER                13
#define TIMER_MIN 0
#define TIMER_MAX 13
/* message tags */
#define MSG_RQSTVEC 1
#define MSG_SENDVEC 2
#define MSG_END 3


/*----------------------------*/
/* TYPEDEFs                   */
/*----------------------------*/
/* dummy buffer */
typedef int MSGDUMMY, *PMSGDUMMY;

/*----------------------------*/
/* CONSTANTs                  */
/*----------------------------*/
static const char szFileIn[FILENAM_SIZE] = "lap.in";

/*----------------------------*/
/* VARIABLEs                  */
/*----------------------------*/
/* output files */
static USHORT fOutMode;     /* flag word for output modes 
                             0 0 0 0  (bit mask)
                             | | | |
                             | | | -----> output psi
                             | | -------> output final solution
                             | ---------> output error
                             -----------> output timers 
                            */
static char szFileOutPsi[FILENAM_SIZE];
static char szFileOutSol[FILENAM_SIZE];
static char szFileOutErr[FILENAM_SIZE];
static char szFileOutTime[FILENAM_SIZE];
#ifdef DEF_OUTMSG
static char szFileOutMsg[FILENAM_SIZE];
#endif
/* input files */
static char szFileInSol[FILENAM_SIZE];
/* current processor number */
static ULONG ulProc;
/* Gaussian-Legendre Quadrature */
static USHORT usQuadN;
static FPTYPE *afpQuadX;
static FPTYPE *afpQuadW;
static FPTYPE *afpQuadXt;   /* scaled */
static FPTYPE *afpQuadWt;   /* scaled */
/* Numerical Data */
static USHORT usPolyN;
static USHORT usTessN;
static ULONG ulIterN; 
static ULONG aulBufPrio[NP_][N_]; /* priority to buffer submatrix */
static ULONG ulBufPrio1N;         /* number of elements that are 1 */
/* Stored Element Data */
static FPTYPE afpStoreP1[M_][3];
static FPTYPE afpStoreP2[M_][3];
static FPTYPE afpStoreP3[M_][3];
static FPTYPE afpStoreU1[M_][3];
static FPTYPE afpStoreU2[M_][3];
static FPTYPE afpStoreX[M_][3];
static FPTYPE afpStoreN[M_][3];
static FPTYPE afpStoreDA[M_];
static FPTYPE afpStoreJ[M_];
static FPTYPE fpStoreA;
/* Superimposed Electric Field */
static FPTYPE afpE0[3];
/* Geometrical Data of Particles */
static FPTYPE afpZ[N_][3];   /* center */
static FPTYPE afpR[N_];      /* radius */
static FPTYPE afpQ[N_];      /* charge */
/* system matrix A */
static FPTYPE afpA[NP_][NBUF_][M_][M_];
/* reference to buffered submatrices in afpA */
static ULONG aulABufRef[NP_][N_];
/* RHS vector b */
static FPTYPE afpB[NMP_];
/* range of particles on processor */
static ULONG ulParticleStart, ulParticleStop;
/* range of rows on processor */
static ULONG ulRowStart, ulRowStop;
/* surface potential */
static FPTYPE *afpPsi;
/* local sub solution vector */
static FPTYPE *afpX;
/* exact solution vector */
static FPTYPE *afpXSol;
/* error in each iteration */
static FPTYPE *afpXErr;
/* test file */
static FILE *pFileTest;

/*----------------------------*/
/* PROTOTYPEs                 */
/*----------------------------*/
static void InitializeTimers(void);
static void OutputTimers(void);
static void OutputTimer(FILE *pFileOutPsi, int iTimer, char szName[20]);
static void Initialize(void);
static void CleanUp(void);
static void StoreElements(void);
static void AdjustGeometry(void);
static void CreateMatrices(void);
static void InitializeSolve(void);
static void Solve(void);
static void OutputPsi(void);
static void OutputSol(void);
static void OutputErr(void);
static void CreateAnn(ULONG ulN, FPTYPE *pfpA, ULONG ulIncr);
static void CreateAkl(ULONG ulK, ULONG ulL, FPTYPE *pfpA, ULONG ulIncr);
static void CreateBn(ULONG ulK, FPTYPE *pfpB);

/*----------------------------*/
/* FUNCTION main              */
/*----------------------------*/
int main(void)
{
  USHORT usM;      /* number of elements on one particle */
 
  /* initialize timers */
  InitializeTimers();
  CMMD_node_timer_start(TIMER_TOTAL);
  /* initialize program */
  CMMD_node_timer_start(TIMER_INITIALIZE);
  Initialize();
  if (ulProc == 0)
  {
    printf("--- CreateGeometry ---\n"); 
  }
  if ((usM = CreateGeometry(usPolyN, usTessN)) == 0)
  {
    fprintf(stderr, "GEOMETRY - not enough memory\n");
    CleanUp();
    return -1;
  }
  if (usM != M_)
  {
    fprintf(stderr, "NSPHER - invalid number of elements\n");
    CleanUp();
    return -1;
  }
  CMMD_node_timer_stop(TIMER_INITIALIZE);
  CMMD_node_timer_start(TIMER_STOREELEMENTS);
  if (ulProc == 0)
  {
    printf("--- StoreElements ---\n"); 
  }
  StoreElements();
  CMMD_node_timer_stop(TIMER_STOREELEMENTS);
  CMMD_node_timer_start(TIMER_CLOSEGEOMETRY);
  if (ulProc == 0)
  {
    printf("--- CloseGeometry ---\n"); 
  }
  CloseGeometry();
  CMMD_node_timer_stop(TIMER_CLOSEGEOMETRY);
  CMMD_node_timer_start(TIMER_ADJUSTGEOMETRY);
  if (ulProc == 0)
  {
    printf("--- AdjustGeometry ---\n"); 
  }
  AdjustGeometry();
  CMMD_node_timer_stop(TIMER_ADJUSTGEOMETRY);
  CMMD_node_timer_start(TIMER_CREATEMATRICES);

  if (ulProc == 0) printf("--- CreateMatrices ---\n");
/*
#ifdef WWT
  {
    int t;
    t = wwt_get_vt_lo();
    printf("PE %d: Time before CreateMatrices = %d\n", ulProc, t);
  }
#endif
*/
  CreateMatrices();

  CMMD_node_timer_stop(TIMER_CREATEMATRICES);
  CMMD_node_timer_start(TIMER_INITIALIZESOLVE);
  if (ulProc == 0)
  {
    printf("--- InitializeSolve ---\n"); 
  }
  InitializeSolve();
  CMMD_node_timer_stop(TIMER_INITIALIZESOLVE);
  CMMD_node_timer_start(TIMER_SOLVE);
  if (ulProc == 0)
  {
    printf("--- Solve ---\n"); 
  }
  Solve();
  CMMD_node_timer_stop(TIMER_SOLVE);
  CMMD_node_timer_start(TIMER_OUTPUT);
  if (fOutMode & 1)
  {
    if (ulProc == 0)
    {
      printf("--- OutputPsi ---\n"); 
    }
    OutputPsi();
  }

  if (fOutMode & 2)
  {
    if (ulProc == 0)
    {
      printf("--- OutputSol ---\n"); 
    }
    OutputSol();
  }
  if (fOutMode & 4)
  {
    if (ulProc == 0)
    {
      printf("--- OutputErr ---\n"); 
    }
    OutputErr();
  }
  CMMD_node_timer_stop(TIMER_OUTPUT);
  CMMD_node_timer_start(TIMER_CLEANUP);

  if (ulProc == 0)
  {
    printf("--- CleanUp ---\n"); 
  }
  CleanUp();
  CMMD_node_timer_stop(TIMER_CLEANUP);
  CMMD_node_timer_stop(TIMER_TOTAL);
  if (ulProc == 0) {
      printf("Total time %f seconds\n", CMMD_node_timer_elapsed(TIMER_TOTAL));
      }
  if (fOutMode & 8)
  {
    if (ulProc == 0)
    {
      printf("--- OutputTimers ---\n");
    }
    OutputTimers();
  }
#ifdef WWT
  printf("PE %d: about to return.\n", ulProc);
#endif
  return 0;
}

/*----------------------------*/
/* FUNCTION InitializeTimers  */
/*----------------------------*/
static void InitializeTimers(void)
{
  int iTimer;
  
  /* initialize all timers */
  for (iTimer = TIMER_MIN; iTimer <= TIMER_MAX; iTimer++)
  {
    CMMD_node_timer_clear(iTimer);
  }
}

/*----------------------------*/
/* FUNCTION OutputTimers      */
/*----------------------------*/
static void OutputTimers(void)
{
  FILE *pFileOutTime;

  /* open output file for append */
  pFileOutTime = fopen(szFileOutTime, "a");
  fprintf(pFileOutTime, "#\n# --- Timers ---\n");
  fprintf(pFileOutTime, 
          "#                           min                   max\n");
  OutputTimer(pFileOutTime, TIMER_TOTAL, "total");
  OutputTimer(pFileOutTime, TIMER_INITIALIZE, "Initialize");
  OutputTimer(pFileOutTime, TIMER_STOREELEMENTS, "StoreElements");
  OutputTimer(pFileOutTime, TIMER_CLOSEGEOMETRY, "CloseGeometry");
  OutputTimer(pFileOutTime, TIMER_ADJUSTGEOMETRY, "AdjustGeometry");
  OutputTimer(pFileOutTime, TIMER_CREATEMATRICES, "CreateMatrices");
  OutputTimer(pFileOutTime, TIMER_INITIALIZESOLVE, "InitializeSolve");
  OutputTimer(pFileOutTime, TIMER_SOLVE, "Solve");
  OutputTimer(pFileOutTime, TIMER_COMMUNICATE, "Communicate");
  OutputTimer(pFileOutTime, TIMER_RECREATEMATRICES, "RecreateMatrices");
  OutputTimer(pFileOutTime, TIMER_BARRIER, "Barrier");
  OutputTimer(pFileOutTime, TIMER_OUTPUT, "Output");
  OutputTimer(pFileOutTime, TIMER_CLEANUP, "CleanUp");
  /* close output file */
  fclose(pFileOutTime);
}

/*----------------------------*/
/* FUNCTION OutputTimer       */
/*----------------------------*/
static void OutputTimer(FILE *pFileOutPsi, int iTimer, char szName[20])
{
  double fpBusy, fpBusyMin, fpBusyMax;
#ifndef WWT
  fpBusy = CMMD_node_timer_busy(iTimer); 
  fpBusyMin = CMMD_reduce_double(fpBusy, CMMD_combiner_dmin);
  fpBusyMax = CMMD_reduce_double(fpBusy, CMMD_combiner_dmax);
  fprintf(pFileOutPsi, "# %-20s %21.12g %21.12g\n", szName, fpBusyMin,
          fpBusyMax);
#endif
}

/*----------------------------*/
/* FUNCTION Initialize        */
/*----------------------------*/
static void Initialize(void)
{
  char szFileQuad[FILENAM_SIZE];
  char szFileNum[FILENAM_SIZE];
  char szFileGeo[FILENAM_SIZE];
  char sz[FILENAM_SIZE * 3];
  FILE *pFileIn;
  FILE *pFileQuad;
  FILE *pFileNum;
  FILE *pFileGeo;
  USHORT usI, usJ;
  ULONG ulI, ulJ, ulK;
  FPTYPE fpXtMax;
  ULONG ulBufPrio;

  /* get current processor number */
  ulProc = (ULONG)CMMD_self_address();

  /* range of particles, rows on this processor */
  ulParticleStart = ulProc * NP_;
  ulParticleStop = (ulProc + 1) * NP_;  
  ulRowStart = ulParticleStart * M_;
  ulRowStop = ulParticleStop * M_;

#ifndef WWT
  /* set general CM open mode */
  CMMD_set_open_mode(CMMD_sync_bc);
  /* set open mode for stdout */
  CMMD_fset_io_mode(stdout, CMMD_independent);
  /* set open mode for stderr */
  CMMD_fset_io_mode(stderr, CMMD_independent);
#endif
#ifndef WWT 
  /* open test file */
  pFileTest = fopen("lap.test", "w");
  CMMD_fset_io_mode(pFileTest, CMMD_sync_seq);
  fprintf(pFileTest, "\n--- Processor %u ---\n", ulProc); 
#endif
  /* at first no output mode */
  fOutMode = 0;
  /* open general input file */
  pFileIn = fopen(szFileIn, "r");
  /* get file names */
  fgets(szFileQuad, sizeof(szFileQuad), pFileIn);
  szFileQuad[strlen(szFileQuad) - 1] = '\0';
  fgets(szFileNum, sizeof(szFileNum), pFileIn);
  szFileNum[strlen(szFileNum) - 1] = '\0';
  fgets(szFileGeo, sizeof(szFileGeo), pFileIn);
  szFileGeo[strlen(szFileGeo) - 1] = '\0';
  fgets(szFileOutPsi, sizeof(szFileOutPsi), pFileIn);
  szFileOutPsi[strlen(szFileOutPsi) - 1] = '\0';
  if (strlen(szFileOutPsi) > 0)
  {
    /* output surface potentials for each iteration */
    fOutMode |= 1;
  }
  fgets(szFileOutSol, sizeof(szFileOutSol), pFileIn);
  szFileOutSol[strlen(szFileOutSol) - 1] = '\0';
  if (strlen(szFileOutSol) > 0)
  {
    /* output final phi vector */ 
    fOutMode |= 2;
  }
  fgets(sz, sizeof(sz), pFileIn);
  sz[strlen(sz) - 1] = '\0';
  if (strlen(sz) > 0)
  {
    /* output error for each iteration */
    fOutMode |= 4;
    sscanf(sz, "%s %s", szFileOutErr, szFileInSol);
  }
  fgets(szFileOutTime, sizeof(szFileOutTime), pFileIn);
  szFileOutTime[strlen(szFileOutTime) - 1] = '\0';
  if (strlen(szFileOutTime) > 0)
  {
    /* output times */ 
    fOutMode |= 8;
  }
#ifdef DEF_OUTMSG
  fgets(szFileOutMsg, sizeof(szFileOutMsg), pFileIn);
  szFileOutMsg[strlen(szFileOutMsg) - 1] = '\0';
#endif
  /* close general input file */
  fclose(pFileIn);

  
  /* read input file for Gaussian-Legendre Quadrature parameters */
  pFileQuad = fopen(szFileQuad, "r");
  fscanf(pFileQuad, "%hu", &usQuadN);
  /* allocate memory for data */
  afpQuadX = (FPTYPE *)malloc(usQuadN * sizeof(FPTYPE));
  afpQuadW = (FPTYPE *)malloc(usQuadN * sizeof(FPTYPE));
  for (usI = 0; usI < usQuadN; usI++)
  {
    fscanf(pFileQuad, "%lf %lf", &afpQuadX[usI], &afpQuadW[usI]);
  }
  fclose(pFileQuad);
  /* calculate scaled parameters */
  afpQuadXt = (FPTYPE *)malloc(usQuadN * usQuadN * sizeof(FPTYPE));
  afpQuadWt = (FPTYPE *)malloc(usQuadN * usQuadN * sizeof(FPTYPE));
  for (usI = 0; usI < usQuadN; usI++)
  {
    fpXtMax = 1. - afpQuadX[usI];
    for (usJ = 0; usJ < usQuadN; usJ++)
    {
      *(afpQuadXt+usI*usQuadN+usJ) = afpQuadX[usJ] * fpXtMax;
      *(afpQuadWt+usI*usQuadN+usJ) = afpQuadW[usJ] * fpXtMax;
    }
  }

  /* read input file for numerical data */
  pFileNum = fopen(szFileNum, "r");
  fscanf(pFileNum, "%hu", &usPolyN);
  fscanf(pFileNum, "%hu", &usTessN);
  fscanf(pFileNum, "%u", &ulIterN);
  /* read priority to buffer submatrix for each particle-particle pair */
  ulBufPrio1N = 0;
  for (ulK = 0, ulI = 0; ulI < N_; ulI++)
  {
    if (ulI >= ulParticleStart && ulI < ulParticleStop)
    {
      for (ulJ = 0; ulJ < N_; ulJ++)
      {
        fscanf(pFileNum, "%u", &ulBufPrio);
        aulBufPrio[ulK][ulJ] = ulBufPrio;
        if (ulBufPrio == 1)
        {
          /* count number of highest priorities */
          ulBufPrio1N++;
        }
      }
      ulK++;
    }
    else
    {
      for (ulJ = 0; ulJ < N_; ulJ++)
      {
        fscanf(pFileNum, "%u", &ulBufPrio);
      }
    }
  }
  fclose(pFileNum);
  
  /* read input file for geometrical data */
  pFileGeo = fopen(szFileGeo, "r");
  fscanf(pFileGeo, "%lf %lf %lf", 
         &afpE0[0], &afpE0[1], &afpE0[2]);
  for (ulI = 0; ulI < N_; ulI++)
  {
    fscanf(pFileGeo, "%lf %lf %lf %lf %lf", 
           &afpZ[ulI][0], &afpZ[ulI][1], &afpZ[ulI][2], 
           &afpR[ulI], &afpQ[ulI]);
  }
  fclose(pFileGeo);
}

/*----------------------------*/
/* FUNCTION StoreElements     */
/*----------------------------*/
void StoreElements(void)
{
  GEO geo;
  USHORT usI;

  fpStoreA = 0.;
  for (usI = 0; usI < M_; usI++)
  {
    GetElementAll(usI, &geo);
    afpStoreP1[usI][0] = geo.afpP1[0];
    afpStoreP1[usI][1] = geo.afpP1[1];
    afpStoreP1[usI][2] = geo.afpP1[2];
    afpStoreP2[usI][0] = geo.afpP2[0];
    afpStoreP2[usI][1] = geo.afpP2[1];
    afpStoreP2[usI][2] = geo.afpP2[2];
    afpStoreP3[usI][0] = geo.afpP3[0];
    afpStoreP3[usI][1] = geo.afpP3[1];
    afpStoreP3[usI][2] = geo.afpP3[2];
    afpStoreU1[usI][0] = geo.afpU1[0];
    afpStoreU1[usI][1] = geo.afpU1[1];
    afpStoreU1[usI][2] = geo.afpU1[2];
    afpStoreU2[usI][0] = geo.afpU2[0];
    afpStoreU2[usI][1] = geo.afpU2[1];
    afpStoreU2[usI][2] = geo.afpU2[2];
    afpStoreX[usI][0] = geo.afpX[0];
    afpStoreX[usI][1] = geo.afpX[1];
    afpStoreX[usI][2] = geo.afpX[2];
    afpStoreN[usI][0] = geo.afpN[0];
    afpStoreN[usI][1] = geo.afpN[1];
    afpStoreN[usI][2] = geo.afpN[2];
    fpStoreA += afpStoreDA[usI] = geo.fpDA;
    afpStoreJ[usI] = geo.fpJacob;
  }
}
  
/*----------------------------*/
/* FUNCTION AdjustGeometry    */
/*----------------------------*/
void AdjustGeometry(void)
{
  USHORT usI;
  FPTYPE fpRadiusPolyInv;

  fpRadiusPolyInv = sqrt((4. * PI) / fpStoreA);
  for (usI = 0; usI < M_; usI++)
  {
    afpStoreP1[usI][0] *= fpRadiusPolyInv; 
    afpStoreP1[usI][1] *= fpRadiusPolyInv;  
    afpStoreP1[usI][2] *= fpRadiusPolyInv;
    afpStoreP2[usI][0] *= fpRadiusPolyInv; 
    afpStoreP2[usI][1] *= fpRadiusPolyInv;
    afpStoreP2[usI][2] *= fpRadiusPolyInv;
    afpStoreP3[usI][0] *= fpRadiusPolyInv; 
    afpStoreP3[usI][1] *= fpRadiusPolyInv; 
    afpStoreP3[usI][2] *= fpRadiusPolyInv; 
    afpStoreU1[usI][0] *= fpRadiusPolyInv;
    afpStoreU1[usI][1] *= fpRadiusPolyInv;
    afpStoreU1[usI][2] *= fpRadiusPolyInv;
    afpStoreU2[usI][0] *= fpRadiusPolyInv;
    afpStoreU2[usI][1] *= fpRadiusPolyInv;
    afpStoreU2[usI][2] *= fpRadiusPolyInv;
    afpStoreX[usI][0] *= fpRadiusPolyInv;
    afpStoreX[usI][1] *= fpRadiusPolyInv;
    afpStoreX[usI][2] *= fpRadiusPolyInv;
    afpStoreDA[usI] *= fpRadiusPolyInv * fpRadiusPolyInv;
    afpStoreJ[usI] *= fpRadiusPolyInv * fpRadiusPolyInv;
  }
  fpStoreA *= fpRadiusPolyInv * fpRadiusPolyInv;
}

/*----------------------------*/
/* FUNCTION InitializeSolve   */
/*----------------------------*/
static void InitializeSolve(void)
{
  FILE *pFileInSol;
  ULONG ulI, ulII;
  FPTYPE fpXSol;

  /* allocate memory for local solution vector */
  afpX = (FPTYPE *)malloc(NMP_ * sizeof(FPTYPE));
  if (fOutMode & 1)
  {
    /* allocate memory for surface potential */
    afpPsi = (FPTYPE *)malloc(ulIterN * NP_ * sizeof(FPTYPE));
  }
  if (fOutMode & 4)
  {
    /* allocate memory for final solution vector */
    afpXSol = (FPTYPE *)malloc(NMP_ * sizeof(FPTYPE));
    /* allocate memory for error */
    afpXErr = (FPTYPE *)malloc(ulIterN * sizeof(FPTYPE));
    /* read it from file */
    pFileInSol = fopen(szFileInSol, "r");
    ulII = 0;
    for (ulI = 0; ulI < NM_; ulI++)
    {
      fscanf(pFileInSol, "%lf", &fpXSol);
      if (ulI >= ulRowStart && ulI < ulRowStop)
      {
        afpXSol[ulII] = fpXSol;
        ulII++;
      }
    }
    fclose(pFileInSol);
  }
}


/*----------------------------*/
/* FUNCTION CleanUp           */
/*----------------------------*/
static void CleanUp(void)
{
  /* free memory */
  free(afpQuadX);
  free(afpQuadW);
  free(afpQuadXt);
  free(afpQuadWt);
  free(afpX);
  if (fOutMode & 1)
  {
    free(afpPsi);
  }
  if (fOutMode & 4)
  {
    free(afpXSol);
    free(afpXErr);
  }
  /* close test file */
#ifndef WWT
  fclose(pFileTest);
#endif
}
 
/*----------------------------*/
/* FUNCTION CreateMatrices    */
/*----------------------------*/
static void CreateMatrices(void)
{
  ULONG ulI, ulJ, ulII;
  ULONG ulK;
  ULONG ulBufPrioMin;
  ULONG ulJMin;

  /* loop through rows of particles */
  for (ulII = 0, ulI = ulParticleStart; ulI < ulParticleStop; ulII++, ulI++)
  {
    /* initialize aulABufRef */
    for (ulJ = 0; ulJ < N_; ulJ++)
    {
      aulABufRef[ulII][ulJ] = N_;
    }
    /* loop through buffered submatrices */
    for (ulK = 0; ulK < NBUF_; ulK++)
    {
      /* initialize ulBufPrioMin with highest possible priority */
      ulBufPrioMin = (ULONG)N_ * (ULONG)N_;
      /* search for submatrices with highest priority and buffer these */
      for (ulJ = 0; ulJ < N_; ulJ++)
      {
        /* if submatrix is not buffered yet, and its priority is greater
           than the priorities of all other submatrices */
        if (aulABufRef[ulII][ulJ] == N_ &&
            aulBufPrio[ulII][ulJ] <= ulBufPrioMin)
        {
          ulBufPrioMin = aulBufPrio[ulII][ulJ];
          ulJMin = ulJ;
        }
      }
      aulABufRef[ulII][ulJMin] = ulK;
      /* check if diagonal block matrix */
      if (ulI == ulJMin)
      {
        /* diagonal block matrix */
        CreateAnn(ulI, &afpA[ulII][ulK][0][0], M_);
      }
      else
      {
        /* off-diagonal block matrix */
        CreateAkl(ulI, ulJMin, &afpA[ulII][ulK][0][0], M_);
      }
    }
    /* create RHS subvector */
    CreateBn(ulI, &afpB[ulII * M_]);
  }
}
                   
/*-------------------------------------*/
/* FUNCTION CreateAnn                  */
/* ulN - particle number               */
/* pfpA - position of submatrix        */
/* ulIncr - increment from row to row  */
/*-------------------------------------*/
static void CreateAnn(ULONG ulN, FPTYPE *pfpA, ULONG ulIncr)
{
  static ULONG ulI, ulJ;
  static USHORT usIXs, usIXt;
  static FPTYPE afpXmY[3];
  static FPTYPE fpR;
  static FPTYPE fpRxy2, fpRxy;
  static FPTYPE fpSum;
  static FPTYPE afpY[3];
  static FPTYPE *pfpAA, *pfpAAA;

  /* radius of particle ulN */
  fpR = afpR[ulN];  

  /* loop through rows */ 
  for (ulI = 0, pfpAA = pfpA; ulI < M_; ulI++, pfpAA += ulIncr)
  {
    /* loop through columns */
    for (ulJ = 0, pfpAAA = pfpAA;
         ulJ < M_; 
         ulJ++, pfpAAA++)
    {
      /* check if diagonal element */
      if (ulJ == ulI)
      {
        *pfpAAA = -afpStoreDA[ulI] / fpStoreA;
      } /* check if diagonal element */
      else
      {
        afpXmY[0] = (afpStoreX[ulI][0] - afpStoreX[ulJ][0]) *
                    fpR;
        afpXmY[1] = (afpStoreX[ulI][1] - afpStoreX[ulJ][1]) *
                    fpR;
        afpXmY[2] = (afpStoreX[ulI][2] - afpStoreX[ulJ][2]) *
                    fpR;
        fpRxy2 = afpXmY[0]*afpXmY[0] + afpXmY[1]*afpXmY[1] +
                  afpXmY[2]*afpXmY[2];
        /* check if large distance */
        if (fpRxy2 > 18. * afpStoreDA[ulJ] * fpR * fpR)
        {
          /* no Gaussian-Legendre Quadrature */
          fpRxy = sqrt(fpRxy2);
          *pfpAAA = -afpStoreDA[ulJ] * fpR * fpR *
                  .5 * (afpStoreN[ulJ][0]*afpXmY[0] + 
                  afpStoreN[ulJ][1]*afpXmY[1] +
                  afpStoreN[ulJ][2]*afpXmY[2]) /
                  (PI * fpRxy * fpRxy * fpRxy) -
                  afpStoreDA[ulJ] / fpStoreA;
        } /* check if large distance */
        else
        {
          /* Gaussian-Legendre Quadrature */
          fpSum = 0.;
          for (usIXs = 0; usIXs < usQuadN; usIXs++)
          {
            for (usIXt = 0; usIXt < usQuadN; usIXt++)
            {
              afpY[0] = afpQuadX[usIXs] * afpStoreU1[ulJ][0] + 
                        *(afpQuadXt+usIXs*usQuadN+usIXt) *
                        afpStoreU2[ulJ][0] + 
                        afpStoreP2[ulJ][0];
              afpY[1] = afpQuadX[usIXs] * afpStoreU1[ulJ][1] + 
                        *(afpQuadXt+usIXs*usQuadN+usIXt) *
                        afpStoreU2[ulJ][1] + 
                        afpStoreP2[ulJ][1];
              afpY[2] = afpQuadX[usIXs] * afpStoreU1[ulJ][2] + 
                        *(afpQuadXt+usIXs*usQuadN+usIXt) *
                        afpStoreU2[ulJ][2] + 
                        afpStoreP2[ulJ][2];
              afpXmY[0] = (afpStoreX[ulI][0] - afpY[0]) * fpR;
              afpXmY[1] = (afpStoreX[ulI][1] - afpY[1]) * fpR;
              afpXmY[2] = (afpStoreX[ulI][2] - afpY[2]) * fpR;
              fpRxy = sqrt(afpXmY[0]*afpXmY[0] +
                           afpXmY[1]*afpXmY[1] +
                           afpXmY[2]*afpXmY[2]);
              fpSum += afpQuadW[usIXs] *  
                       *(afpQuadWt+usIXs*usQuadN+usIXt) *
                       afpStoreJ[ulJ] *
                       (afpStoreN[ulJ][0]*afpXmY[0] + 
                       afpStoreN[ulJ][1]*afpXmY[1] +
                       afpStoreN[ulJ][2]*afpXmY[2]) /
                       (fpRxy * fpRxy * fpRxy);
            }
          }
          *pfpAAA = -fpSum * .5 * fpR * fpR / PI -
                   afpStoreDA[ulJ] / fpStoreA;
        } /* check if large distance */
      } /* check if diagonal element */
    } /* loop through columns */ 
  } /* loop through rows */ 
}

/*-------------------------------------*/
/* FUNCTION CreateAkl                  */
/* ulK - particle number (x)           */
/* ulL - particle number (y)           */
/* pfpA - position of submatrix        */
/* ulIncr - increment from row to row  */
/*-------------------------------------*/
static void CreateAkl(ULONG ulK, ULONG ulL, FPTYPE *pfpA, ULONG ulIncr)
{
  static ULONG ulI, ulJ;
  static USHORT usIXs, usIXt;
  static FPTYPE afpXmY[3];
  static FPTYPE fpRxy2, fpRxy;
  static FPTYPE fpSum;
  static FPTYPE afpXX[3];
  static FPTYPE afpY[3];
  static FPTYPE *pfpAA, *pfpAAA;
  static FPTYPE afpZdst[3];

  /* distance between particle centers */
  afpZdst[0] = afpZ[ulK][0] - afpZ[ulL][0];
  afpZdst[1] = afpZ[ulK][1] - afpZ[ulL][1];
  afpZdst[2] = afpZ[ulK][2] - afpZ[ulL][2];

  /* loop through rows */ 
  for (ulI = 0, pfpAA = pfpA; ulI < M_; ulI++, pfpAA += ulIncr)
  { 
    afpXX[0] = afpStoreX[ulI][0] * afpR[ulK];
    afpXX[1] = afpStoreX[ulI][1] * afpR[ulK];
    afpXX[2] = afpStoreX[ulI][2] * afpR[ulK];
    /* loop through columns */
    for (ulJ = 0, pfpAAA = pfpAA;
         ulJ < M_; 
         ulJ++, pfpAAA++)
    {
      afpXmY[0] = afpXX[0] -
                  afpStoreX[ulJ][0] * afpR[ulL] +  
                  afpZdst[0]; 
      afpXmY[1] = afpXX[1] -
                  afpStoreX[ulJ][1] * afpR[ulL] +  
                  afpZdst[1]; 
      afpXmY[2] = afpXX[2] -
                  afpStoreX[ulJ][2] * afpR[ulL] +  
                  afpZdst[2]; 
      fpRxy2 = afpXmY[0]*afpXmY[0] + afpXmY[1]*afpXmY[1] +
               afpXmY[2]*afpXmY[2];
      /* check if large distance */
      if (fpRxy2 > 18. * afpStoreDA[ulJ] * afpR[ulL] * afpR[ulL])
      {
        /* no Gaussian-Legendre Quadrature */
        fpRxy = sqrt(fpRxy2);
        *pfpAAA = -afpStoreDA[ulJ] * afpR[ulL] * afpR[ulL] *
                .5 * (afpStoreN[ulJ][0]*afpXmY[0] + 
                afpStoreN[ulJ][1]*afpXmY[1] +
                afpStoreN[ulJ][2]*afpXmY[2]) /
                (PI * fpRxy * fpRxy * fpRxy);
      } /* check if large distance */
      else
      {
        /* Gaussian-Legendre Quadrature */
        fpSum = 0.;
        for (usIXs = 0; usIXs < usQuadN; usIXs++)
        {
          for (usIXt = 0; usIXt < usQuadN; usIXt++)
          {
            afpY[0] = afpQuadX[usIXs] * afpStoreU1[ulJ][0] + 
                      *(afpQuadXt+usIXs*usQuadN+usIXt) *
                      afpStoreU2[ulJ][0] + 
                      afpStoreP2[ulJ][0];
            afpY[1] = afpQuadX[usIXs] * afpStoreU1[ulJ][1] + 
                      *(afpQuadXt+usIXs*usQuadN+usIXt) *
                      afpStoreU2[ulJ][1] + 
                      afpStoreP2[ulJ][1];
            afpY[2] = afpQuadX[usIXs] * afpStoreU1[ulJ][2] + 
                      *(afpQuadXt+usIXs*usQuadN+usIXt) *
                      afpStoreU2[ulJ][2] + 
                      afpStoreP2[ulJ][2];
            afpXmY[0] = afpXX[0] -
                        afpY[0] * afpR[ulL] +  
                        afpZdst[0]; 
            afpXmY[1] = afpXX[1] -
                        afpY[1] * afpR[ulL] +  
                        afpZdst[1]; 
            afpXmY[2] = afpXX[2] -
                        afpY[2] * afpR[ulL] +  
                        afpZdst[2]; 
            fpRxy = sqrt(afpXmY[0]*afpXmY[0] +
                         afpXmY[1]*afpXmY[1] +
                         afpXmY[2]*afpXmY[2]);
            fpSum += afpQuadW[usIXs] *  
                     *(afpQuadWt+usIXs*usQuadN+usIXt) *
                     afpStoreJ[ulJ] *
                     (afpStoreN[ulJ][0]*afpXmY[0] + 
                     afpStoreN[ulJ][1]*afpXmY[1] +
                     afpStoreN[ulJ][2]*afpXmY[2]) /
                     (fpRxy * fpRxy * fpRxy);
          }
        }
        *pfpAAA = -fpSum * .5 * afpR[ulL] * afpR[ulL] / PI;
      } /* check if large distance */
    } /* loop through columns */ 
  } /* loop through rows */ 
}

/*-------------------------------------*/
/* FUNCTION CreateBn                   */
/* ulN - particle number               */
/* pfpB - position of subvector        */
/*-------------------------------------*/
static void CreateBn(ULONG ulK, FPTYPE *pfpB)
{
  static ULONG ulI;
  static ULONG ulL;
  static FPTYPE fpSum;
  static FPTYPE afpXX[3];
  static FPTYPE afpXmZ[3];
  static FPTYPE fpRxz;
  static FPTYPE *pfpBB;

  /* loop through rows */
  for (ulI = 0, pfpBB = pfpB; ulI < M_; ulI++, pfpBB++)
  {
    afpXX[0] = afpStoreX[ulI][0] * afpR[ulK] + afpZ[ulK][0];
    afpXX[1] = afpStoreX[ulI][1] * afpR[ulK] + afpZ[ulK][1];
    afpXX[2] = afpStoreX[ulI][2] * afpR[ulK] + afpZ[ulK][2];
    fpSum = 0.;
    /* loop through particles */
    for (ulL = 0; ulL < N_; ulL++)
    {
      afpXmZ[0] = afpXX[0] - afpZ[ulL][0]; 
      afpXmZ[1] = afpXX[1] - afpZ[ulL][1]; 
      afpXmZ[2] = afpXX[2] - afpZ[ulL][2]; 
      fpRxz = sqrt(afpXmZ[0] * afpXmZ[0] +
                   afpXmZ[1] * afpXmZ[1] +
                   afpXmZ[2] * afpXmZ[2]);
      fpSum += afpQ[ulL] / fpRxz;
    }
    *pfpBB = -fpSum + afpE0[0] * afpXX[0] + afpE0[1] * afpXX[1] +
            afpE0[2] * afpXX[2];
  }
}

/*-------------------------------------*/
/* FUNCTION Solve                      */
/*-------------------------------------*/
static void Solve(void) 
{
  ULONG ulI, ulII, ulJ, ulJJ, ulK, ulKK, ulL;
  ULONG ulStart, ulStop;
  ULONG ulP;
  ULONG ulABuf;
  ULONG ulIter;
  FPTYPE fpSum;
  CMMD_mcb mcb;
  CMMD_mcb amcbRec[P_];
  CMMD_mcb amcbSend[SENDMSG_MAX];
  ULONG ulSendN;
  MSGDUMMY msgdummy;
  USHORT fTest;
  USHORT afReceived[P_];
  ULONG ulMsgTag;
  ULONG ulPSource;
  FPTYPE afpATemp[M_][M_];    /* non buffered submatrix */
  FPTYPE afpXOld[NM_];
  FPTYPE *afpXBuf;
  FPTYPE *pfpPsi;
  FPTYPE *pfpXBuf, *pfpXBuf1;
  FPTYPE *pfpXErr;
  FPTYPE *pfpA;
#ifdef DEF_OUTMSG
  FILE *pFileMsg;
#endif

  /* allocate memory for buffer solution matrix */
  afpXBuf = (FPTYPE *)malloc(NMP_ * N_ * sizeof(FPTYPE));

  for (ulI = 0; ulI < NMP_ * N_; ulI++)
    afpXBuf[ulI] = 0.;

  /* initial guess */
  for (ulII = 0; ulII < NMP_; ulII++)
  {
    afpX[ulII] = afpB[ulII];
  }
    
  /* if output of psi */
  if (fOutMode & 1)
  {
    /* pfpPsi points to first element in afpPsi */
    pfpPsi = afpPsi;
  }
  /* if output of solution error */
  if (fOutMode & 4)
  {
    /* pfpXErr points to first element in afpXErr */
    pfpXErr = afpXErr;
  }
  /* if output of messages */
#ifdef DEF_OUTMSG
    /* open temporary message file in synchronous sequential mode */
    pFileMsg = fopen(szFileOutMsg, "w");
    CMMD_fset_io_mode(pFileMsg, CMMD_sync_seq);
    /* start timer */
    CMMD_node_timer_start(TIMER_MSG);
#endif
/*
#ifdef WWT
  {
    int t;
    t = wwt_get_vt_lo();
    printf("PE %d: Time before BARRIER = %d\n", ulProc, t);
  }
#endif
*/
  /* synchronize all nodes */
  /* CMMD_sync_with_nodes(); */

  /* iteration */ 
  for (ulIter = 0; ulIter < ulIterN; ulIter++)
  {
    if (ulProc == 0)
      {
	printf("Iteration %u\n", ulIter);
      }
    /* if output of messages */
#ifdef DEF_OUTMSG
      CMMD_node_timer_stop(TIMER_MSG);
      fprintf(pFileMsg, "I%3u %3u %21.12g\n", ulIter, ulProc, 
              CMMD_node_timer_busy(TIMER_MSG));
      CMMD_node_timer_start(TIMER_MSG);
#endif
    /* store requests to transmit sub vectors and to create sub
       matrices */
    /* loop thru columns (processors) */
    for (ulP = 0; ulP < P_; ulP++)
    {
      fTest = 0;
      /* loop thru cols (particles) */
      ulStart = ulP * NP_;
      ulStop = ulStart + (ULONG)NP_;
      for (ulL = ulStart; ulL < ulStop; ulL++)
      {
        /* loop thru rows (particles) */
        for (ulKK = 0; ulKK < NP_; ulKK++)
        {
          /* if sub vector of proc ulP is to be received */
          if (ulIter % aulBufPrio[ulKK][ulL] == 0)
          {
            fTest = 1;
            break;
          }
        }
      }
      /* if sub vector of proc ulP is to be received */
      if (fTest == 1)
      {
        /* if vector is on different proc */
        if (ulP != ulProc)
        {
          /* send request to send sub solution vector */
          CMMD_node_timer_start(TIMER_COMMUNICATE);
          mcb = CMMD_send_async(ulP, MSG_RQSTVEC, 
                                &msgdummy, sizeof(msgdummy), NULL, 0); 
          CMMD_free_mcb(mcb);
          /* receive sub solution vector */
          amcbRec[ulP] = CMMD_receive_async(ulP, MSG_SENDVEC, 
                                &afpXOld[ulP * NMP_], 
                                NMP_*sizeof(FPTYPE), NULL, 0); 
          CMMD_node_timer_stop(TIMER_COMMUNICATE);
          afReceived[ulP] = 0;
          /* if output of messages */
#ifdef DEF_OUTMSG
            fprintf(pFileMsg, "R%3u\n", ulP);
#endif
        } 
        else
        {
          /* copy new x to old x within current proc */
          for (ulII = 0, ulI = ulP * NMP_; ulII < NMP_; ulII++, ulI++)
          {
            afpXOld[ulI] = afpX[ulII];
          }
        } /* if vector is on different proc */
      } /* if sub vector of proc ulP is to be received */
    } /* loop thru columns (processors) */
    /* loop thru columns */
#if 0
    for (ulII = 0; ulII < NMP_; ulII++)
    {
      afpX[ulII] = afpB[ulII];
    }
#endif
    /* so far no data sent */
    ulSendN = 0;
    /* pfpXBuf points to first element in afpXBuf */
    pfpXBuf = afpXBuf;

    /* loop through rows (particles) */
    for (ulKK = 0, ulK = ulParticleStart; ulKK < NP_; ulKK++, ulK++)
    {
      /* top and bottom row of particle */
      ulStart = ulKK * M_;
      ulStop = ulStart + (ULONG)M_;
      /* loop through columns (particles) */
      for (ulL = 0; ulL < N_; ulL++)
      {
#if 0
        /* if priority of submatrix is 1 */
        if (aulBufPrio[ulKK][ulL] == 1)
        {
          /* do not store sub solution vector in afpXBuf but in afpXBuf1 */
          pfpXBuf1 = afpXBuf1;
        }
        else
        {
          pfpXBuf1 = pfpXBuf;
          /* step to next buffered element */
          pfpXBuf += M_;
        }
#endif
        /* if sub solution vector is to be updated */
        if (ulIter % aulBufPrio[ulKK][ulL] == 0)
        {
          ulABuf = aulABufRef[ulKK][ulL]; 
          /* if submatrix is not buffered */
          if (ulABuf == N_)
          {
            CMMD_node_timer_start(TIMER_RECREATEMATRICES);
            /* create submatrix */
            /* check if diagonal block matrix */
            if (ulK == ulL)
            {
              /* diagonal block matrix */
              CreateAnn(ulK, &afpATemp[0][0], M_);
            }
            else
            {
              /* off-diagonal block matrix */
              CreateAkl(ulK, ulL, &afpATemp[0][0], M_);
            }
	    pfpA = &(afpATemp[0][0]);
            CMMD_node_timer_stop(TIMER_RECREATEMATRICES);
          } /* if submatrix is not buffered */
	  else
	    {
	      pfpA = &(afpA[ulKK][ulABuf][0][0]);
	    }
          /* proc from which to get sub vector */
          ulP = ulL / (ULONG)NP_;
          /* if vector is on different proc */
          if (ulP != ulProc)
          {
            /* if sub vector is not received yet */
            if (!afReceived[ulP])
            {
              CMMD_node_timer_start(TIMER_COMMUNICATE);
              /* (synchronize) wait until async receive is done */
              do 
              {
                /* message loop: look if there are any requests while  
                                 waiting for async receive */
                while (ulSendN < SENDMSG_MAX) 
                {
                  /* are there any messages for current proc? */
                  mcb = CMMD_mcb_pending(CMMD_ANY_NODE, MSG_RQSTVEC);
                  if (mcb != NULL)
                  {
                    ulPSource = CMMD_mcb_node(mcb);
                    CMMD_receive_async(ulPSource, MSG_RQSTVEC,
                                       &msgdummy, sizeof(msgdummy), NULL, 0);
                    CMMD_msg_wait(mcb);
                    CMMD_free_mcb(mcb);
                    amcbSend[ulSendN] = CMMD_send_async(ulPSource, MSG_SENDVEC, 
                                  &afpXOld[ulRowStart], 
                                  NMP_*sizeof(FPTYPE), NULL, 0); 
                    ulSendN++;
                    /* if output of messages */
#ifdef DEF_OUTMSG
                      fprintf(pFileMsg, "S%3u\n", ulPSource);
#endif
                  } 
                  else
                  {
                    break;
                  }
                } /* message loop */
              } while (CMMD_msg_done(amcbRec[ulP]) == FALSE); /* sync */
              /* free mcb */
              CMMD_free_mcb(amcbRec[ulP]);
              CMMD_node_timer_stop(TIMER_COMMUNICATE);
              afReceived[ulP] = 1;
            }
          } /* if vector is on different proc */
          /* store product of submatrix and subvector */
	  /* loop thru rows */
	  for (ulI = 0, pfpXBuf1 = pfpXBuf, ulII = ulStart; 
	       ulI < M_; 
	       ulI++, pfpXBuf1++, ulII++)
	    {
	      fpSum = 0.;
	      /* loop thru cols */
	      for (ulJJ = 0, ulJ = ulL * M_; ulJJ < M_; ulJJ++, ulJ++)
		{
		  fpSum += *(pfpA++) * afpXOld[ulJ];
		}
	      afpX[ulII] += fpSum - *pfpXBuf1;
	      *pfpXBuf1 = fpSum;
	    }
        } /* if sub solution vector is to be updated */
        else
        {
          /* second message loop because first message loop is
             not sufficient, if there are no requests from 
             current proc, since then first msg loop is not executed */
          while (ulSendN < SENDMSG_MAX) 
          {
            /* are there any messages for current proc? */
            mcb = CMMD_mcb_pending(CMMD_ANY_NODE, MSG_RQSTVEC);
            if (mcb != NULL)
            {
              ulPSource = CMMD_mcb_node(mcb);
              CMMD_receive_async(ulPSource, MSG_RQSTVEC,
                                 &msgdummy, sizeof(msgdummy), NULL, 0);
              CMMD_msg_wait(mcb);
              CMMD_free_mcb(mcb);
              amcbSend[ulSendN] = CMMD_send_async(ulPSource, MSG_SENDVEC, 
                            &afpXOld[ulRowStart], 
                            NMP_*sizeof(FPTYPE), NULL, 0); 
              ulSendN++;
              /* if output of messages */
#ifdef DEF_OUTMSG
                fprintf(pFileMsg, "S%3u\n", ulPSource);
#endif
            } 
            else
            {
              break;
            }
          } /* message loop */
        } /* if sub solution vector is to be updated */
        
	pfpXBuf += M_;

      } /* loop through columns (particles) */

      /* if output of psi */
      if (fOutMode & 1)
	{
	  /* compute surface potentials */
        fpSum = 0.;
        for (ulII = ulStart, ulI = 0; ulII < ulStop; ulII++, ulI++)
	  {
	    fpSum += afpX[ulII] * afpStoreDA[ulI];      
        }
        *pfpPsi = -fpSum / fpStoreA;
        pfpPsi++;
      }
    } /* loop through rows (particles) */
    /* if output of solution error */
    if (fOutMode & 4)
    {
      /* compute error on current processor */
      fpSum = 0.;
      for (ulII = 0; ulII < NMP_; ulII++)
      {
        fpSum += fabs((afpX[ulII] - afpXSol[ulII]) / afpXSol[ulII]);
      }
      *pfpXErr = fpSum;
      pfpXErr++;
    }
    CMMD_node_timer_start(TIMER_COMMUNICATE);
    /* wait for all messages to be sent */
    for (ulI = 0; ulI < ulSendN; ulI++)
    {
      CMMD_msg_wait(amcbSend[ulI]);
      CMMD_free_mcb(amcbSend[ulI]);
    }
    CMMD_node_timer_stop(TIMER_COMMUNICATE);
#ifdef DEF_OUTMSG
      fprintf(pFileMsg, "E%3u\n", ulIter); 
#endif

      if (ulProc == 0) printf("End of iteration %d\n", ulIter);
  } /* iteration */

  CMMD_node_timer_start(TIMER_BARRIER);
  /* send MSG_END to each other processor */
  for (ulP = 0; ulP < P_; ulP++)
  {
    if (ulP != ulProc)
    {
      mcb = CMMD_send_async(ulP, MSG_END, 
                            &msgdummy, sizeof(msgdummy), NULL, 0);
      CMMD_free_mcb(mcb);
    }
  }
 
  /* wait for all other processors to finish and
     respond to requests while waiting */
  ulP = 0;
  /* loop until all processors have sent MSG_END */
  while (ulP < (ULONG)(P_ - 1))
  {
    /* message loop */
    while ((mcb = CMMD_mcb_pending(CMMD_ANY_NODE, CMMD_ANY_TAG)) == NULL);
    ulPSource = CMMD_mcb_node(mcb);
    ulMsgTag = CMMD_mcb_tag(mcb);
    CMMD_receive_async(ulPSource, ulMsgTag,
                       &msgdummy, sizeof(msgdummy), NULL, 0);
    CMMD_msg_wait(mcb);
    CMMD_free_mcb(mcb);
    switch (ulMsgTag)
    {
      case MSG_RQSTVEC:
        /* send partial solution vector */
        mcb = CMMD_send_async(ulPSource, MSG_SENDVEC, 
                        afpX, NMP_*sizeof(FPTYPE), NULL, 0); 
        CMMD_free_mcb(mcb);
        break;
      case MSG_END:
        /* count processors which have finished */
        ulP++;
        break;
      default:
        break;
    } 
  } /* loop until all processors have sent MSG_END */ 
  CMMD_node_timer_stop(TIMER_BARRIER);

  CMMD_sync_with_nodes(); /* is this needed? 7/10 */

  /* free memory */
  free(afpXBuf);


  /* if output of messages */
#ifdef DEF_OUTMSG
    fclose(pFileMsg);
    /* stop timer */
    CMMD_node_timer_stop(TIMER_MSG);
#endif
}

/*-------------------------------------*/
/* FUNCTION OutputPsi                  */
/*-------------------------------------*/
static void OutputPsi(void)
{
  FILE *pFileOutPsi;
  ULONG ulIter, ulK;
  FPTYPE *pfpPsi;

#ifndef WWT
  pFileOutPsi = fopen(szFileOutPsi, "w");
  CMMD_fset_io_mode(pFileOutPsi, CMMD_sync_seq);

  /* only proc 0 prints out header */
  if (ulProc == 0)
  {
    fprintf(pFileOutPsi, "# Particles = %lu\n", N_);
    fprintf(pFileOutPsi, "# buffered particles = %lu\n", NBUF_);
    fprintf(pFileOutPsi, "# Elements/Particle = %lu\n", M_);
    fprintf(pFileOutPsi, "floating-point precision = %u\n", sizeof(FPTYPE));
    fprintf(pFileOutPsi, "E0 = (%21.12g,%21.12g,%21.12g)\n",
          afpE0[0], afpE0[1], afpE0[2]);
  }
  /* init pointer to psi array */
  pfpPsi = afpPsi;
  /* iteration */
  for (ulIter = 0; ulIter < ulIterN; ulIter++)
  {
    if (ulProc == 0) 
    {
      fprintf(pFileOutPsi, "\n--- Iteration %3hu ---\n", ulIter + 1);
      fprintf(pFileOutPsi, 
        "Sphere     Q       X       Y       Z       R              Psi\n");
    }
    for (ulK = ulParticleStart; ulK < ulParticleStop; ulK++)
    { 
      fprintf(pFileOutPsi, 
              "%4lu %7.3g %7.3g %7.3g %7.3g %7.3g %21.12g\n",
              ulK, afpQ[ulK], afpZ[ulK][0], afpZ[ulK][1], 
              afpZ[ulK][2], afpR[ulK], *pfpPsi);
      pfpPsi++;
    }
    /* flush buffer */
    fflush(pFileOutPsi);
  }
  /* close output file */
  fclose(pFileOutPsi);

 cant_do_outputpsi:
#endif
}

/*-------------------------------------*/
/* FUNCTION OutputSol                  */
/*-------------------------------------*/
static void OutputSol(void)
{
  FILE *pFileOutSol;
  ULONG ulII;
#ifndef WWT
  pFileOutSol = fopen(szFileOutSol, "w");

  CMMD_fset_io_mode(pFileOutSol, CMMD_sync_seq);
  /* output x on every element on current proc */
  for (ulII = 0; ulII < NMP_; ulII++)
  {
    fprintf(pFileOutSol, "%28.22g\n", afpX[ulII]);
  }
#else
  /* in WWT only P0 prints its stuff */
  pFileOutSol = fopen(szFileOutSol, "w");
  
  if (ulProc == 0)
    for (ulII = 0; ulII < NMP_; ulII++)
      {
	fprintf(pFileOutSol, "%28.22g\n", afpX[ulII]);
      } 

#endif
  /* only last proc prints out header */
  if (ulProc == P_ - 1)
  {
    fprintf(pFileOutSol, "\n-----------------------\n");
    fprintf(pFileOutSol, "# Particles = %lu\n", N_);
    fprintf(pFileOutSol, "# buffered particles = %lu\n", NBUF_);
    fprintf(pFileOutSol, "# Elements/Particle = %lu\n", M_);
    fprintf(pFileOutSol, "floating-point precision = %u\n", sizeof(FPTYPE));
    fprintf(pFileOutSol, "E0 = (%21.12g,%21.12g,%21.12g)\n",
          afpE0[0], afpE0[1], afpE0[2]);
  }

  /* close output file */
  fclose(pFileOutSol);

}

/*-------------------------------------*/
/* FUNCTION OutputErr                  */
/*-------------------------------------*/
static void OutputErr(void)
{
  FILE *pFileOutErr;
  ULONG ulIter;
  FPTYPE fpLambda;
  double fpTimeSolve;
  FPTYPE fpTimeIter;
#ifndef WWT
  /* sum error vectors of all nodes */
  if (sizeof(FPTYPE) == sizeof(float))
  {
    CMMD_reduce_v(afpXErr, afpXErr, CMMD_combiner_fadd, sizeof(FPTYPE), 
                  ulIterN);
  }
  else
  {
    CMMD_reduce_v(afpXErr, afpXErr, CMMD_combiner_dadd, sizeof(FPTYPE), 
                  ulIterN);
  }
  /* if timers are created */
  if (fOutMode & 8)
  {
    /* get maximal solution time */ 
    fpTimeSolve = CMMD_node_timer_busy(TIMER_SOLVE); 
    fpTimeIter = (FPTYPE)CMMD_reduce_double(fpTimeSolve, CMMD_combiner_dmax) /
                 (FPTYPE)ulIterN;
  }

  pFileOutErr = fopen(szFileOutErr, "a");
  CMMD_fset_io_mode(pFileOutErr, CMMD_independent);
  /* only proc 0 does output */
  if (ulProc == 0)
  {
    fprintf(pFileOutErr, 
            "# Iter  Error                    Lambda                 Time\n");
    for (ulIter = 0; ulIter < ulIterN; ulIter++)
    {
      afpXErr[ulIter] /= (FPTYPE)NM_;
      fpLambda = pow(afpXErr[ulIter], 1. / (FPTYPE)(ulIter + 2));
      /* if timers created */
      if (fOutMode & 8)
      {
        fprintf(pFileOutErr, "%3u %21.12g %21.12g %21.12g\n", ulIter + 1, 
                afpXErr[ulIter], fpLambda, 
                (FPTYPE)(ulIter + 1) * fpTimeIter);
      }
      else
      {
        fprintf(pFileOutErr, "%3u %21.12g %21.12g\n", ulIter + 1, 
                afpXErr[ulIter], fpLambda);
      }
    }
    fprintf(pFileOutErr, "\n# -----------------------\n");
    fprintf(pFileOutErr, "# # Particles = %lu\n", N_);
    fprintf(pFileOutErr, "# # buffered particles = %lu\n", NBUF_);
    fprintf(pFileOutErr, "# # Elements/Particle = %lu\n", M_);
    fprintf(pFileOutErr, "# floating-point precision = %u\n", sizeof(FPTYPE));
    fprintf(pFileOutErr, "# E0 = (%21.12g,%21.12g,%21.12g)\n",
          afpE0[0], afpE0[1], afpE0[2]);
  }
  /* close output file */
  fclose(pFileOutErr);
#endif
}
